package top.hcode.hoj.controller.admin;

import cn.hutool.core.io.FileUtil;
import cn.hutool.core.map.MapUtil;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import org.apache.shiro.authz.annotation.Logical;
import org.apache.shiro.authz.annotation.RequiresAuthentication;
import org.apache.shiro.authz.annotation.RequiresRoles;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.StringUtils;
import org.springframework.web.bind.annotation.*;
import top.hcode.hoj.common.result.CommonResult;
import top.hcode.hoj.pojo.dto.AnnouncementDto;
import top.hcode.hoj.pojo.dto.ProblemDto;
import top.hcode.hoj.pojo.entity.*;
import top.hcode.hoj.pojo.vo.AnnouncementVo;
import top.hcode.hoj.service.impl.*;
import top.hcode.hoj.utils.Constants;

import javax.validation.Valid;
import java.io.File;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;


/**
 * @Author: Himit_ZH
 * @Date: 2020/12/19 22:28
 * @Description:
 */
@RestController
@RequestMapping("/api/admin/contest")
public class AdminContestController {

    @Autowired
    private ContestServiceImpl contestService;

    @Autowired
    private ProblemServiceImpl problemService;

    @Autowired
    private ContestProblemServiceImpl contestProblemService;

    @Autowired
    private ContestAnnouncementServiceImpl contestAnnouncementService;

    @Autowired
    private AnnouncementServiceImpl announcementService;

    @GetMapping("/get-contest-list")
    @RequiresAuthentication
    @RequiresRoles(value = {"root", "admin"}, logical = Logical.OR)
    public CommonResult getContestList(@RequestParam(value = "limit", required = false) Integer limit,
                                       @RequestParam(value = "currentPage", required = false) Integer currentPage,
                                       @RequestParam(value = "keyword", required = false) String keyword) {

        if (currentPage == null || currentPage < 1) currentPage = 1;
        if (limit == null || limit < 1) limit = 10;
        IPage<Contest> iPage = new Page<>(currentPage, limit);
        QueryWrapper<Contest> queryWrapper = new QueryWrapper<>();

        if (!StringUtils.isEmpty(keyword)) {
            keyword = keyword.trim();
            queryWrapper
                    .like("title", keyword).or()
                    .like("id", keyword);
        }

        queryWrapper.orderByDesc("start_time");

        IPage<Contest> contestList = contestService.page(iPage, queryWrapper);
        if (contestList.getTotal() == 0) { // 未查询到一条数据
            return CommonResult.successResponse(contestList, "暂无数据");
        } else {
            return CommonResult.successResponse(contestList, "获取成功");
        }
    }

    @GetMapping("")
    @RequiresAuthentication
    @RequiresRoles(value = {"root", "admin"}, logical = Logical.OR)
    public CommonResult getContest(@Valid @RequestParam("cid") Long cid) {
        Contest contest = contestService.getById(cid);
        if (contest != null) { // 查询成功
            return CommonResult.successResponse(contest, "查询成功！");
        } else {
            return CommonResult.errorResponse("查询失败！", CommonResult.STATUS_FAIL);
        }
    }

    @DeleteMapping("")
    @RequiresAuthentication
    @RequiresRoles(value = "root")
    public CommonResult deleteContest(@Valid @RequestParam("cid") Long cid) {
        boolean result = contestService.removeById(cid);
        /*
        contest的id为其他表的外键的表中的对应数据都会被一起删除！
         */
        if (result) { // 删除成功
            return CommonResult.successResponse(null, "删除成功！");
        } else {
            return CommonResult.errorResponse("删除失败！", CommonResult.STATUS_FAIL);
        }
    }

    @PostMapping("")
    @RequiresAuthentication
    @RequiresRoles(value = {"root", "admin"}, logical = Logical.OR)
    public CommonResult addContest(@RequestBody Contest contest) {
        boolean result = contestService.save(contest);
        if (result) { // 添加成功
            return CommonResult.successResponse(null, "添加成功！");
        } else {
            return CommonResult.errorResponse("添加失败", CommonResult.STATUS_FAIL);
        }
    }

    @PutMapping("")
    @RequiresAuthentication
    @RequiresRoles(value = {"root", "admin"}, logical = Logical.OR)
    public CommonResult updateContest(@RequestBody Contest contest) {
        boolean result = contestService.saveOrUpdate(contest);
        if (result) { // 添加成功
            return CommonResult.successResponse(null, "修改成功！");
        } else {
            return CommonResult.errorResponse("修改失败", CommonResult.STATUS_FAIL);
        }
    }

    @PutMapping("/change-contest-visible")
    @RequiresAuthentication
    @RequiresRoles(value = {"root", "admin"}, logical = Logical.OR)
    public CommonResult changeContestVisible(@RequestParam(value = "cid", required = true) Long cid,
                                             @RequestParam(value = "visible", required = true) Boolean visible) {

        boolean result = contestService.saveOrUpdate(new Contest().setId(cid).setVisible(visible));
        if (result) { // 添加成功
            return CommonResult.successResponse(null, "修改成功！");
        } else {
            return CommonResult.errorResponse("修改失败", CommonResult.STATUS_FAIL);
        }
    }

    /**
     * 以下为比赛的题目的增删改查操作接口
     */

    @GetMapping("/get-problem-list")
    @RequiresAuthentication
    @RequiresRoles(value = {"root", "admin"}, logical = Logical.OR)
    @Transactional
    public CommonResult getProblemList(@RequestParam(value = "limit", required = false) Integer limit,
                                       @RequestParam(value = "currentPage", required = false) Integer currentPage,
                                       @RequestParam(value = "keyword", required = false) String keyword,
                                       @RequestParam(value = "cid", required = false) Long cid,
                                       @RequestParam(value = "problemType", required = false) Integer problemType) {
        if (cid == null) {
            return CommonResult.errorResponse("参数错误！", CommonResult.STATUS_NOT_FOUND);
        }

        if (currentPage == null || currentPage < 1) currentPage = 1;
        if (limit == null || limit < 1) limit = 10;
        IPage<Problem> iPage = new Page<>(currentPage, limit);
        // 根据cid在ContestProblem表中查询到对应pid集合
        QueryWrapper<ContestProblem> contestProblemQueryWrapper = new QueryWrapper<>();
        contestProblemQueryWrapper.select("pid").eq("cid", cid);
        List<Long> pidList = new LinkedList<>();
        contestProblemService.list(contestProblemQueryWrapper).forEach(contestProblem -> {
            pidList.add(contestProblem.getPid());
        });

        if (pidList.size() == 0 && problemType == null) { // 该比赛原本就无题目数据
            return CommonResult.successResponse(iPage, "获取成功");
        }

        QueryWrapper<Problem> problemQueryWrapper = new QueryWrapper<>();

        if (problemType != null) { // 必备条件 私有的不可取来做比赛题目
            problemQueryWrapper.eq("type", problemType).ne("auth", 2);
        }

        // 逻辑判断，如果是查询已有的就应该是in，如果是查询不要重复的，使用not in
        if (pidList.size() > 0) {
            if (problemType != null) {
                // 同时需要与比赛相同类型的题目，权限需要是公开的（私有，比赛中不可加入！）
                problemQueryWrapper.notIn("id", pidList);
            } else {
                problemQueryWrapper.in("id", pidList);
            }
        }

        if (!StringUtils.isEmpty(keyword)) {
            problemQueryWrapper
                    .like("title", keyword).or()
                    .like("problem_id", keyword).or()
                    .like("author", keyword);
        }

        IPage<Problem> problemList = problemService.page(iPage, problemQueryWrapper);
        if (problemList.getTotal() == 0) { // 未查询到一条数据
            return CommonResult.successResponse(problemList, "暂无数据");
        } else {
            return CommonResult.successResponse(problemList, "获取成功");
        }
    }

    @GetMapping("/problem")
    @RequiresAuthentication
    @RequiresRoles(value = {"root", "admin"}, logical = Logical.OR)
    public CommonResult getProblem(@Valid @RequestParam("pid") Long pid) {
        Problem problem = problemService.getById(pid);
        if (problem != null) { // 查询成功
            return CommonResult.successResponse(problem, "查询成功！");
        } else {
            return CommonResult.errorResponse("查询失败！", CommonResult.STATUS_FAIL);
        }
    }

    @DeleteMapping("/problem")
    @RequiresAuthentication
    @RequiresRoles(value = {"root", "admin"}, logical = Logical.OR)
    public CommonResult deleteProblem(@Valid @RequestParam("pid") Long pid) {
        boolean result = problemService.removeById(pid);
        /*
        problem的id为其他表的外键的表中的对应数据都会被一起删除！
         */
        if (result) { // 删除成功
            FileUtil.del(Constants.File.TESTCASE_BASE_FOLDER.getPath() + File.separator + "problem_" + pid);
            return CommonResult.successResponse(null, "删除成功！");
        } else {
            return CommonResult.errorResponse("删除失败！", CommonResult.STATUS_FAIL);
        }
    }

    @PostMapping("/problem")
    @RequiresAuthentication
    @RequiresRoles(value = {"root", "admin"}, logical = Logical.OR)
    @Transactional
    public CommonResult addProblem(@RequestBody ProblemDto problemDto) {

        QueryWrapper<Problem> queryWrapper = new QueryWrapper<>();
        queryWrapper.eq("problem_id", problemDto.getProblem().getProblemId().toUpperCase());
        Problem problem = problemService.getOne(queryWrapper);
        if (problem != null) {
            return CommonResult.errorResponse("该题目的Problem ID已存在，请更换！", CommonResult.STATUS_FAIL);
        }

        boolean result = problemService.adminAddProblem(problemDto);
        if (result) { // 添加成功
            // 顺便返回新的题目id，好下一步添加外键操作
            return CommonResult.successResponse(MapUtil.builder().put("pid", problemDto.getProblem().getId()).map(), "添加成功！");
        } else {
            return CommonResult.errorResponse("添加失败", CommonResult.STATUS_FAIL);
        }

    }

    @PutMapping("/problem")
    @RequiresAuthentication
    @RequiresRoles(value = {"root", "admin"}, logical = Logical.OR)
    @Transactional
    public CommonResult updateProblem(@RequestBody ProblemDto problemDto) {

        QueryWrapper<Problem> queryWrapper = new QueryWrapper<>();
        queryWrapper.eq("problem_id", problemDto.getProblem().getProblemId().toUpperCase());
        Problem problem = problemService.getOne(queryWrapper);

        // 如果problem_id不是原来的且已存在该problem_id，则修改失败！
        if (problem != null && problem.getId().longValue() != problemDto.getProblem().getId()) {
            return CommonResult.errorResponse("当前的Problem ID 已被使用，请重新更换新的！", CommonResult.STATUS_FAIL);
        }

        boolean result = problemService.adminUpdateProblem(problemDto);
        if (result) { // 更新成功
            return CommonResult.successResponse(null, "修改成功！");
        } else {
            return CommonResult.errorResponse("修改失败", CommonResult.STATUS_FAIL);
        }
    }

    @GetMapping("/contest-problem")
    @RequiresAuthentication
    @RequiresRoles(value = {"root", "admin"}, logical = Logical.OR)
    public CommonResult getContestProblem(@RequestParam(value = "cid", required = true) Long cid,
                                          @RequestParam(value = "pid", required = true) Long pid) {
        QueryWrapper<ContestProblem> queryWrapper = new QueryWrapper<>();
        queryWrapper.eq("cid", cid).eq("pid", pid);
        ContestProblem contestProblem = contestProblemService.getOne(queryWrapper);
        if (contestProblem != null) {
            return CommonResult.successResponse(contestProblem, "查询成功！");
        } else {
            return CommonResult.errorResponse("查询失败", CommonResult.STATUS_FAIL);
        }
    }

    @PutMapping("/contest-problem")
    @RequiresAuthentication
    @RequiresRoles(value = {"root", "admin"}, logical = Logical.OR)
    public CommonResult setContestProblem(@RequestBody ContestProblem contestProblem) {
        boolean result = contestProblemService.saveOrUpdate(contestProblem);
        if (result) {
            return CommonResult.successResponse(contestProblem, "更新成功！");
        } else {
            return CommonResult.errorResponse("更新失败", CommonResult.STATUS_FAIL);
        }
    }

    @PutMapping("/change-problem-auth")
    @RequiresAuthentication
    @RequiresRoles(value = {"root", "admin"}, logical = Logical.OR)
    public CommonResult changeProblemAuth(@RequestBody Problem problem) {
        boolean result = problemService.saveOrUpdate(problem);
        if (result) { // 更新成功
            return CommonResult.successResponse(null, "修改成功！");
        } else {
            return CommonResult.errorResponse("修改失败", CommonResult.STATUS_FAIL);
        }
    }

    @PostMapping("/add-problem-from-public")
    @RequiresAuthentication
    @RequiresRoles(value = {"root", "admin"}, logical = Logical.OR)
    public CommonResult addProblemFromPublic(@RequestBody HashMap<String, String> params) {

        String pidStr = params.get("pid");
        String cidStr = params.get("cid");
        String displayId = params.get("displayId");
        if (StringUtils.isEmpty(pidStr) || StringUtils.isEmpty(cidStr) || StringUtils.isEmpty(displayId)) {
            return CommonResult.errorResponse("参数错误！", CommonResult.STATUS_FAIL);
        }

        Long pid = Long.valueOf(pidStr);
        Long cid = Long.valueOf(cidStr);

        QueryWrapper<ContestProblem> contestProblemQueryWrapper = new QueryWrapper<>();
        contestProblemQueryWrapper.eq("pid", pid).eq("cid", cid);
        ContestProblem contestProblem = contestProblemService.getOne(contestProblemQueryWrapper, false);
        if (contestProblem != null) {
            return CommonResult.errorResponse("添加失败，该题目已经加入到此次比赛当中了，请勿重复操作！", CommonResult.STATUS_FAIL);
        }

        // 比赛中题目显示默认为原标题
        String displayName = problemService.getById(pid).getTitle();

        boolean result = contestProblemService.saveOrUpdate(new ContestProblem()
                .setCid(cid).setPid(pid).setDisplayTitle(displayName).setDisplayId(displayId));
        if (result) { // 添加成功
            return CommonResult.successResponse(null, "添加成功！");
        } else {
            return CommonResult.errorResponse("添加失败", CommonResult.STATUS_FAIL);
        }
    }


    /**
     * 以下处理比赛公告的操作请求
     */

    @GetMapping("/announcement")
    @RequiresAuthentication
    @RequiresRoles(value = {"root", "admin"}, logical = Logical.OR)
    public CommonResult getAnnouncementList(@RequestParam(value = "limit", required = false) Integer limit,
                                            @RequestParam(value = "currentPage", required = false) Integer currentPage,
                                            @RequestParam(value = "cid", required = true) Long cid) {
        if (currentPage == null || currentPage < 1) currentPage = 1;
        if (limit == null || limit < 1) limit = 10;

        IPage<AnnouncementVo> announcementList = announcementService.getContestAnnouncement(cid, false, limit, currentPage);
        if (announcementList.getTotal() == 0) { // 未查询到一条数据
            return CommonResult.successResponse(announcementList, "暂无数据");
        } else {
            return CommonResult.successResponse(announcementList, "获取成功");
        }
    }

    @DeleteMapping("/announcement")
    @RequiresAuthentication
    @RequiresRoles(value = {"root", "admin"}, logical = Logical.OR)
    public CommonResult deleteAnnouncement(@Valid @RequestParam("aid") Long aid) {
        boolean result = announcementService.removeById(aid);
        if (result) { // 删除成功
            return CommonResult.successResponse(null, "删除成功！");
        } else {
            return CommonResult.errorResponse("删除失败！", CommonResult.STATUS_FAIL);
        }
    }

    @PostMapping("/announcement")
    @RequiresAuthentication
    @RequiresRoles(value = {"root", "admin"}, logical = Logical.OR)
    @Transactional
    public CommonResult addAnnouncement(@RequestBody AnnouncementDto announcementDto) {
        boolean result1 = announcementService.save(announcementDto.getAnnouncement());
        boolean result2 = contestAnnouncementService.saveOrUpdate(new ContestAnnouncement().setAid(announcementDto.getAnnouncement().getId())
                .setCid(announcementDto.getCid()));
        if (result1 && result2) { // 添加成功
            return CommonResult.successResponse(null, "添加成功！");
        } else {
            return CommonResult.errorResponse("添加失败", CommonResult.STATUS_FAIL);
        }
    }

    @PutMapping("/announcement")
    @RequiresAuthentication
    @RequiresRoles(value = {"root", "admin"}, logical = Logical.OR)
    public CommonResult updateAnnouncement(@RequestBody AnnouncementDto announcementDto) {
        boolean result = announcementService.saveOrUpdate(announcementDto.getAnnouncement());
        if (result) { // 更新成功
            return CommonResult.successResponse(null, "修改成功！");
        } else {
            return CommonResult.errorResponse("修改失败", CommonResult.STATUS_FAIL);
        }
    }
}